(** Time-stamp: <modified the 04/11/2008 (at 16:57) by Erwan Jahier> *)
 
  
open Predef
open PredefEvalType
open PredefEvalConst
open SyntaxTree
open SyntaxTreeCore
open Eff
open Printf
open Lxm
open Errors


(******************************************************************************)
(* We tabulate the result of the [Eff.val_exp] type ckecking. Note,
   that there is no clash in that table because all [var_exp_eff] are
   srcflagged.
*)

let val_exp_eff_type_tab = Hashtbl.create 0
let tabulate_res val_exp_eff type_eff_list =
  Hashtbl.add val_exp_eff_type_tab val_exp_eff type_eff_list

let (val_exp_eff : Eff.val_exp -> Eff.type_ list) =
  fun vef -> 
    try Hashtbl.find val_exp_eff_type_tab vef
    with _ -> assert false


(******************************************************************************)
let finish_me msg = print_string ("\n\tXXX evalType.ml:"^msg^" ->  finish me!\n")
let rec (f : Eff.id_solver -> Eff.val_exp -> Eff.type_ list) =
  fun id_solver ve ->
    let res = 
      match ve with
        | CallByPosEff ({it=posop; src=lxm}, OperEff args) -> (
            try eval_by_pos_type id_solver posop lxm args
            with EvalType_error msg -> 
              raise (Compile_error(lxm, "type error: "^msg))
          )
        | CallByNameEff ({it=nmop; src=lxm}, nmargs ) ->
            try eval_by_name_type id_solver nmop lxm nmargs
            with EvalConst_error msg ->
              raise (Compile_error(lxm, "\n*** can't eval constant: "^msg))
    in
      tabulate_res ve res;
      res

and (eval_by_pos_type :
       Eff.id_solver -> Eff.by_pos_op -> Lxm.t -> Eff.val_exp list -> Eff.type_ list) =
  fun id_solver posop lxm args ->
    match posop with
      | Eff.Predef (op,sargs) -> (
          let targs = List.map (f id_solver) args in
          let res = PredefEvalType.f op lxm sargs targs in
            (* Handling iterator calls: cf comments in LucDump.mli *)
               (match LicDump.poly_op_find lxm with 
                  | Some (LicDump.OpProfile(op_lti, op_lto)) -> 
                      (* Retrieve the substitution *)
                      let targs = List.flatten targs in
                        (match UnifyType.f (targs@res) (op_lti@op_lto) with
                          | UnifyType.Unif t ->
                              LicDump.tabulate_poly_op lxm (LicDump.Subst(t))
                          | UnifyType.Equal -> ()
                          | UnifyType.Ko _ -> assert false
                        )
                  | None -> ()
                  | _ -> assert false 
               );
            res
        )
      | Eff.CALL node_exp_eff -> 
          let lti = List.map (fun v -> v.var_type_eff) node_exp_eff.it.inlist_eff in
          let lto = List.map (fun v -> v.var_type_eff) node_exp_eff.it.outlist_eff in
          let t_args = List.flatten (List.map (f id_solver) args) in
          let llti = List.length lti and lt_args = List.length t_args in
            if llti <> lt_args then
              raise (EvalType_error(
                       sprintf 
                         "\n*** arity error: %d argument(s) are expected, whereas %d is/are provided"
                         llti lt_args))
            else
              (match UnifyType.f lti t_args with
                 | UnifyType.Equal -> lto
                 | UnifyType.Unif subst -> List.map (subst_type subst) lto
                 | UnifyType.Ko msg -> raise (Compile_error(lxm, msg))
              )
                
      | Eff.CONST (id,_) -> [Eff.type_of_const (id_solver.id2const id lxm)]
      | Eff.IDENT id  -> (
          (* [id] migth be a constant, but also a variable *)
          try [Eff.type_of_const (id_solver.id2const id lxm)]
          with _ -> [(id_solver.id2var id lxm).var_type_eff]
        )       
      | Eff.WITH(ve) -> f id_solver ve

      | Eff.TUPLE -> List.flatten (List.map (f id_solver) args)

      | Eff.CONCAT -> (
          match List.map (f id_solver) args with
            | [[Array_type_eff (teff0, size0)]; [Array_type_eff (teff1, size1)]] ->
                let teff = 
                  match UnifyType.f [teff0] [teff1] with
                    | UnifyType.Equal -> teff1
                    | UnifyType.Unif subst -> subst_type subst teff1
                    | UnifyType.Ko msg -> raise (Compile_error(lxm, msg))
                in
                  [Array_type_eff (teff, size0+size1)]
            | _ -> 
                raise(EvalType_error(sprintf "arity error: 2 expected instead of %d" 
                                       (List.length args)))
        )
      | Eff.STRUCT_ACCESS fid -> (
          let type_args_eff = List.flatten (List.map (f id_solver) args) in
            match type_args_eff with 
              | [Struct_type_eff (name, fl)] -> (
                  try [fst (List.assoc fid fl)]
                  with Not_found -> 
                    raise (EvalType_error
                             (Printf.sprintf "%s is not a field of struct %s" 
                                (Ident.to_string fid) 
                                (LicDump.string_of_type_eff(List.hd type_args_eff))))
                )
              | [x] -> type_error [x] "struct type"
              | x -> arity_error x "1"
        )
      | Eff.ARRAY_ACCES  (_, teff) -> [teff] (* XXX check args type! *)

      | Eff.ARRAY_SLICE  (sieff,teff) -> 
          [Array_type_eff(teff, sieff.se_width)]

      | Eff.HAT(size,ceff) -> 
          let teff_list = f id_solver ceff in
            List.map (fun teff -> Array_type_eff(teff, size)) teff_list
              
      | Eff.ARRAY ->
          (* check that args are of the same type *)
          let type_args_eff = (List.map (f id_solver) args) in
          let teff_elt =
            List.fold_left
              (fun acc teff ->
                 match acc with
                   | [] -> teff
                   | [sacc] -> if acc = teff then acc else
                       raise(EvalType_error(
                               "all array elements should be of the same type"))
                   | _ -> assert false
              )
              []
              type_args_eff
          in
            assert (List.length teff_elt = 1);
            [Array_type_eff(List.hd teff_elt, List.length args)]


      | Eff.WHEN clk_exp -> (
          let type_args_eff = List.map (f id_solver) args in
            (match clk_exp with
               | Base -> ()
               | NamedClock( { it = (cc,cv) ; src = lxm }) -> 
                   let vi = id_solver.id2var (Ident.to_idref cv) lxm in
                     (match vi.var_type_eff with
	                | Eff.Bool_type_eff
                        | Eff.Enum_type_eff _ -> ()
	                | teff  -> 
		            let msg = "the type of a clock cannot be " ^
                              (LicDump.string_of_type_eff teff) 
		            in
		              raise(Compile_error(lxm,msg))
                     )
            );
            match type_args_eff with
              | [teff] -> teff
              | _ -> raise(EvalType_error("arity error (1 arg expected)"))
        )
      | Eff.ARROW
      | Eff.FBY -> (
          let type_args_eff = List.map (f id_solver) args in
            match type_args_eff with
              | [init; teff] -> if init = teff then teff else 
                  raise(EvalType_error("type mismatch. "))
              | _ -> raise(EvalType_error("arity error (2 args expected)"))
        )
      | Eff.CURRENT 
      | Eff.PRE -> (
          let type_args_eff = List.map (f id_solver) args in
            match type_args_eff with
              | [teff] -> teff
              | _ -> raise(EvalType_error("arity error (1 arg expected)"))
        )
      | Eff.MERGE _ -> finish_me "merge"; assert false



and (eval_by_name_type : Eff.id_solver -> Eff.by_name_op -> Lxm.t -> 
      (Ident.t Lxm.srcflagged * Eff.val_exp) list -> Eff.type_ list) =
  fun id_solver namop lxm namargs -> 
    match namop with
      | Eff.STRUCT_anonymous -> 
          (* ??? comment faire ici pour recuperer son type ???
             il faut que je recherche  l'aide des noms de champs
             le type structure qui va bien !

             - creer une table [liste des noms de champs -> ident de type structure] ?
             - rajouter dans la table a sa creation une entree dont le nom
             est compos du nom des champs ?
          *)
          finish_me "anonymous struct not yet supported"; 
          assert false
            (*    failwith "Finish me: anonymous struct not yet supported" *)

      | Eff.STRUCT (pn,opid) -> 
          let struct_type = id_solver.id2type opid lxm in
          let _ = 
            match struct_type with
              | Struct_type_eff(sn, fl) -> 
                  List.iter
                    (fun (fn,fv) -> 
                       let fn, lxm = fn.it, fn.src in
                       let (ft,fopt) = 
                         try List.assoc fn fl with Not_found -> 
                           let msg = "type error: bad field" ^ (Ident.to_string fn) in
                             raise (Compile_error(lxm, msg))
                       in
                         (* let's check the type of fv *)
                       let fv_type = f id_solver fv in
                         match UnifyType.f [ft] fv_type with
                           | UnifyType.Unif t -> 
                               (* XXX cannot occur as long as we do not have
                                  user-level polymorphism (really?).*)
                               ()
                           | UnifyType.Equal -> ()
                           | UnifyType.Ko msg -> raise (Compile_error(lxm, msg))
                    )
                    namargs
              | _ -> 
                  raise (Compile_error(lxm, "type error: a structure is expected"))
            
          in
            [struct_type]

