(** Time-stamp: <modified the 11/09/2008 (at 14:49) by Erwan Jahier> *)


open LicDump
open Printf
open Lxm
open Errors
open Eff

let ci2str = LicDump.string_of_clock2

(* exported *)
(* XXX utiliser une map ! *)
type subst1 = (Ident.t * Ident.t) list
type subst2 = (int * Eff.clock) list
type subst = subst1 * subst2

(* exported *)
let (empty_subst:subst) = [],[]

let (add_subst1 : Ident.t -> Ident.t -> subst1 -> subst1) =
  fun id1 id2 s ->
    if List.mem_assoc id1 s then s else (id1,id2)::s

let (add_subst2 : int -> Eff.clock -> subst2 -> subst2) =
  fun i c s ->
    if List.mem_assoc i s then s else (i,c)::s

let (find_subst1 : Ident.t -> subst1 -> Ident.t option) =
  fun id s ->
    try Some (List.assoc id s) with Not_found -> None

let (find_subst2 : int -> subst2 -> Eff.clock option) =
  fun id s ->
    try Some (List.assoc id s) with Not_found -> None



(******************************************************************************)
(* Stuff to generate fresh clock vars *)

let var_type = ref 0
let reset_var_type () = var_type := 0

let (get_var_type : unit -> int) = 
  fun () -> incr var_type; !var_type


(* exported *)
let (get_constant_clock : unit -> Eff.clock) =
  fun () -> ClockVar (get_var_type ())

(******************************************************************************)

let (subst_to_string : subst -> string) = 
  fun (s1,s2) -> 
    let v2s = Ident.to_string in
      (String.concat ", " (List.map (fun (v1,v2) -> (v2s v1) ^ "/" ^ (v2s v2)) s1)) ^
        " + " ^
        (String.concat ", " (List.map (fun (v1,v2) -> (string_of_int v1) ^ "/" ^ 
				         (ci2str v2)) s2))

(* exported *)
let rec (apply_subst:subst -> Eff.id_clock -> Eff.id_clock) =
  fun (s1,s2) (id, c) ->
    match c with
      | BaseEff -> id, BaseEff
      | On((cc,cv),clk) ->
	  let id = match find_subst1 id s1 with Some id2 -> id2 | None -> id in
	  let cv = match find_subst1 cv s1 with Some cv2 -> cv2 | None -> cv in
          let id, clk = apply_subst (s1,s2) (id,clk) in
	    id, On((cc,cv),  clk)
      | ClockVar i ->
          match find_subst2 i s2 with
            | Some clk -> apply_subst (s1,s2) (id,clk)
            | None -> id, c


(* exported *)
(* apply only the second part of the subst *)
let rec (apply_subst2:subst -> Eff.id_clock -> Eff.id_clock) =
  fun (s1,s2) (id,c) ->
      match c with
        | BaseEff -> id, BaseEff 
        | On(v,clk) -> 
            let id, clk = apply_subst2 (s1,s2) (id, clk) in
              id, On(v, clk)
        | ClockVar i -> 
            match find_subst2 i s2 with
              | Some clk -> apply_subst2 (s1,s2) (id,clk)
              | None -> id,c
	    
(* exported *)
let (f : Lxm.t -> subst -> Eff.id_clock -> Eff.id_clock -> subst) =
  fun lxm (s1,s2) ci1 ci2 -> 
    let id1,ci1 = apply_subst (s1,s2) ci1 in
    let id2,ci2 = apply_subst (s1,s2) ci2 in
    let s1 = if id1 = id2 then s1 else add_subst1 id2 id1 s1 in
    let rec aux (s1,s2) (c1,c2) = 
        match (c1,c2) with
	| On(v,clk), BaseEff
	| BaseEff, On(v,clk) -> (s1,s2)
	    (* not unifiable: we stop trying to find a substitution and return s;
	       the error message is issued outside [aux] (just below).
	    *)
	| BaseEff, BaseEff -> (s1,s2) (* ok *)
	| On((cc1,cv1), clk1), On((cc2,cv2), clk2) ->
	    let s1 = if cv1 = cv2 then s1 else add_subst1 cv2 cv1 s1 in
	      aux (s1,s2) (clk1, clk2)
	| ClockVar i, ClockVar j ->
            if i<>j then s1,(add_subst2 i (ClockVar j) s2) else s1,s2
	| ClockVar i, ci
	| ci, ClockVar i -> s1, add_subst2 i ci s2
    in
    let s = aux (s1,s2) (ci1,ci2) in
    let unifiable =
      (* Two clocks are unifiable iff, once s have been applied, they are
         equal modulo the left-most hand-side *)
      match snd (apply_subst s (id1,ci1)), snd (apply_subst s (id2,ci2)) with
	| On(_,c1),On(_,c2) -> c1 = c2
	| BaseEff, BaseEff -> true
	| BaseEff,On(_)
	| On(_),BaseEff -> false    
	| ClockVar _, _
	| _, ClockVar _ -> true
    in
      if unifiable then (
	Verbose.print_string ~level:3 (
	  "# clock checking: unifying '" ^ (ci2str ci1) ^
	    "' and '" ^ (ci2str ci2) ^ "' ->  "^ (subst_to_string s) ^"\n");
	flush stdout;
	s
      ) 
      else
	let msg =  
	  "\n*** clock error: The expression has clock '" ^ 
	    (ci2str ci1) ^ "', \n***             but it is used with clock '" ^ 
	    (ci2str ci2) ^ "'.\n"
	in
	  raise(Compile_error(lxm, msg))

(* exported *)
let (list : Lxm.t -> Eff.id_clock list -> subst -> Eff.id_clock * subst) =
  fun lxm cl s -> 
    List.fold_left 
      (fun (clk,s) clk2 -> 
	 let s = f lxm s clk clk2 in
	 let clk = apply_subst s clk in (* necessary? *)
	   (clk,s)
      )
      (List.hd cl, s)
      (List.tl cl)
